from lark import Tree
from os import listdir
from os.path import join
from swissql.analyzers.rule_checker.custom_rule import CustomRule
from swissql.analyzers.rule_checker.tree_pattern_find import TreePatternFinder
from swissql.utils.exceptions import Error

class RuleFinder:
    SPLITTER: str = ':END_OF_COMMENT:'
    PATTERN: str = 'PATTERN'
    
    rules_dir: str = 'rules'
    rules: list[CustomRule] = []
    rule_files: list[str] = []



    @classmethod
    def __get_rule_names(cls) -> list[str]:
        files = listdir(cls.rules_dir)
        files = list(filter(lambda x: x != '__init__.py', files))
        return files

    @classmethod
    def __parse_rule_file(cls, data: str) -> tuple[str]:
        try:
            comment, grammar = data.split(cls.SPLITTER)
            return (comment, grammar)
        except ValueError as e:
            raise Error('invalid rule file format') from e


    @classmethod
    def initialize(cls):
        cls.rule_files = cls.__get_rule_names()
        

    @classmethod
    def load_rule(cls, rule_file: str):
        data = ''
        with open(join(cls.rules_dir,rule_file), mode='r') as f:
            data = f.read()
        comment, grammar = cls.__parse_rule_file(data)
        cls.rules.append(CustomRule(
            name=rule_file,
            grammar=grammar,
            comment=comment
        ))

    @classmethod
    def load_rules(cls):
        for rule_file in cls.rule_files:
            cls.load_rule(rule_file)            
    
    @classmethod
    def get_rules(cls) -> list[str]:
        return list(cls.rule_files)

    @classmethod
    def find_rules(cls, sql: str) -> list[(CustomRule, list[(int, int)])]:
        found = list()
        for rule in cls.rules:
            try:
                tree = rule.parse_sql_tree(sql)
            except Exception as e:
                continue
            finder = TreePatternFinder()
            finder.visit(tree)
            if len(finder.positions) > 0:
                found.append((rule, finder.positions))
        return found




import sys
import os
from argparse import ArgumentParser, Namespace, BooleanOptionalAction
from swissql.analyzers.style.sql_style_error import SqlfluffCheck
from swissql.analyzers.anti_pattern.anti_pattern_finder import AntiPatternFinder
from swissql.analyzers.syntax.sql_parser import SqlParser
from swissql.analyzers.style.sql_formatter import SqlFormatter
from swissql.utils.schema_reader import SchemaReader
from swissql.utils.exceptions import Error, exception_handler
from swissql.analyzers.optimizations.static_optimizer import StaticOptimizer
from swissql.analyzers.rule_checker.rule_finder import RuleFinder
from swissql.analyzers.rule_checker.custom_rule import CustomRule
from swissql.analyzers.rule_checker.sql_finder import SqlFinder
from swissql.manifest import Manifest


class ArgParser:
    parser: ArgumentParser(prog='swissql')
    args: Namespace
    modes: list[str] = ["syntax", "format", "optimize", "style", "anti_pattern", "rule"]

    @staticmethod
    def __pair_or(argument: str) -> bool:
        return argument not in sys.argv

    @staticmethod
    def __pair_and(argument: str) -> bool:
        return argument in sys.argv

    @classmethod
    def __get_queries(cls) -> list[str]:
        queries = None
        if cls.args.q:
            queries = cls.args.q.split(';')
        elif cls.args.x:
            SqlFinder.initialize()
            if not cls.args.f:
                raise Error('no file provided')
            filename = cls.args.f
            print('\u001b[33m[Extracting Spark SQLs from file using lark]\u001b[0m')
            sqls = SqlFinder.extract_sql_from_file(filename)
            if sqls:
                print('Found:')
                for sql in sqls:
                    print(sql)
            else:
                print('Did not find any SQLs')
            queries = sqls
        elif cls.args.f:
            query = ""
            try:
                with open(cls.args.f, mode="r") as f:
                    query = f.read()
            except FileNotFoundError as e:
                raise Error("query file not found") from e
            queries = query.split(';')
        else:
            raise Error("either -q or -f argument is required")
        queries = list(map(lambda x: x.strip(), queries))
        queries = list(filter(lambda x: x, queries))
        return queries

    @classmethod
    def initialize(cls) -> None:
        RuleFinder.initialize()
        
        rules = RuleFinder.get_rules()       
        optimizers: list[str] = StaticOptimizer.get_optimizers()
        cls.parser = ArgumentParser(
            prog=Manifest.APP_NAME,
            description=Manifest.APP_DESCRIPTION,
        )
        cls.parser.add_argument(
            "mode", choices=cls.modes + ["all", "construct"], help="mode of operation"
        )
        cls.parser.add_argument(
            "-q",
            # "--query",
            required=cls.__pair_or("-f"),
            help="specify SQL query",
        )
        cls.parser.add_argument(
            "-f",
            # "--file-sql",
            required=cls.__pair_or("-q"),
            help="specify file to read SQL query from",
        )
        cls.parser.add_argument(
            "-x",
            # "--extract",
            action=BooleanOptionalAction,
            help="specify if queries should be extracted from source files"
        )
        cls.parser.add_argument(
            "-s",
            # "--schema",
            help="""specify schema:
        Schema is a mapping in one of
            the following forms:\n
                1. {table: {col: type}}\n
                2. {db: {table: {col: type}}}\n
                3. {catalog: {db: {table: {col: type}}}}\n
        Example types: INT, STRING""",
        )
        cls.parser.add_argument(
            "-F",
            # "--File-schema",
            help="specify schema file to read schema from",
        )
        cls.parser.add_argument(
            "-o",
            #  "--optimizer",
            choices=optimizers,
            help="specify optimizers",
        )
        cls.parser.add_argument(
            '-r',
            # "--rule",
            choices=rules+['all',],
            help='specify rule file'
        )
        cls.parser.add_argument(
            '-c',
            # "--choice",
            choices=cls.modes,
            help='specify modes for analysis constructor',
            action='append'
        )
        cls.parser.add_argument(
            "--rules-sqlfluff",
            help="specify rules file for sqlfluff",
        )
        # cls.parser.add_argument("-c", help="specify")
        cls.parser.add_argument(
            "--dialect", default="sparksql", help="specify sqlfluff dialect"
        )
        cls.parser.add_argument(
            "--output-mode",
            default="str",
            choices=["str", "json"],
            help="specify output format",
        )
        cls.args = cls.parser.parse_args()

    @classmethod
    @exception_handler()
    def analyze_one(cls, query: str, mode: str) -> None:
        if mode == "syntax":
            print("\u001b[33m[Generating syntax tree using sqlglot]\u001b[0m")
            output = SqlParser.parse_tree(query)
            print(output)
        elif mode == "format":
            print("\u001b[33m[Formatting sql query using sqlglot]\u001b[0m")
            output = SqlFormatter.format_one(query)
            print(output)
        elif mode == "style":
            print("\u001b[33m[Style sql query use sqlfluff]\u001b[0m")
            # получаем ссылку на файл с помощью lark
            # print(cls.args.f)
            output = SqlfluffCheck(
                "lint",
                file=cls.args.f,
                q=cls.args.q,
                rules=cls.args.rules_sqlfluff,
                dialect=cls.args.dialect,
            ).create_str()
            print(output)
            #
            # Меняем позиции в тексте
        elif mode == "optimize":
            schema = None
            if cls.args.o in ("optimize", "qualify_columns"):
                if cls.args.s:
                    schema = SchemaReader.parse_json_schema(cls.args.s)
                elif cls.args.F:
                    schema = SchemaReader.parse_file_json_schema(cls.args.F)
                else:
                    raise Error("schema is not provided")

            print("\u001b[33m[Optimizing sql query using sqlglot]\u001b[0m")
            print(f"Optimization: {cls.args.o}")
            optimized = StaticOptimizer.optimize(cls.args.o, query, schema)
            print(optimized)

        elif mode == "anti_pattern":
            print("\u001b[33m[Detecting anti-patterns using sqlcheck]\u001b[0m")

            # TODO: add verbosity level option
            instance = AntiPatternFinder(verbose=True)
            sqlcheck_output = ""
            if cls.args.f:
                sqlcheck_output = instance.find_anti_patterns_from_file(cls.args.f)
            elif cls.args.q:
                sqlcheck_output = instance.find_anti_patterns_from_query(cls.args.q)
            else:
                raise Error("either -q or -f argument is required")
            print(sqlcheck_output)
        elif mode == 'rule':
            print('\u001b[33m[Finding rules using lark]\u001b[0m')
            rule = cls.args.r
            if not rule:
                raise Error('no rule specified')
            if rule == 'all':
                RuleFinder.load_rules()
            else:
                RuleFinder.load_rule(rule)
            found = RuleFinder.find_rules(query)
            for composition in found:
                rule, positions = composition
                positions = ' '.join([str(el) for el in positions])
                print(f'Rule {rule.name} found:')
                print(f'Positions: {positions}')
                print(f'Comment: {rule.comment}')
                print()
        elif mode == "construct":
            for mode in cls.args.c:
                cls.analyze_one(query, mode)
                print()
        elif mode == "all":
            for mode in cls.modes:
                cls.analyze_one(query, mode)
                print()

    @classmethod
    def analyze_queries(cls) -> None:
        queries = cls.__get_queries()
        for query in queries:
            print(f'\u001b[33m[Analyzing query:\u001b[35m {query}\u001b[33m]\u001b[37m')
            cls.analyze_one(query, cls.args.mode)
       



from lark import Tree, Visitor

class TreePatternFinder(Visitor):
    positions: list[(int, int)] = []

    def pattern(self, tree):
        self.positions.append((tree.meta.line, tree.meta.column))


class TreeStringFinder(Visitor):
    strings: list[str] = []
    
    def string(self, tree):
        for el in tree.children:
            self.strings.append(el.value)





from enum import Enum
from typing import Union, Callable
from sqlglot import Expression, parse_one
from sqlglot.optimizer import optimize
from sqlglot.optimizer.normalize import normalize
from sqlglot.optimizer.eliminate_subqueries import eliminate_subqueries
from sqlglot.optimizer.expand_multi_table_selects import expand_multi_table_selects
from sqlglot.optimizer.isolate_table_selects import isolate_table_selects
from sqlglot.optimizer.optimize_joins import optimize_joins
from sqlglot.optimizer.pushdown_predicates import pushdown_predicates
from sqlglot.optimizer.pushdown_projections import pushdown_projections
from sqlglot.optimizer.qualify_tables import qualify_tables
from sqlglot.optimizer.qualify_columns import qualify_columns
from sqlglot.optimizer.quote_identities import quote_identities
from sqlglot.optimizer.unnest_subqueries import unnest_subqueries

class StaticOptimizer:
    optimizers = {
        'optimize': optimize,
        'normalize': normalize,
        'eliminate_subqueries': eliminate_subqueries,
        'expand_multi_table_selects': expand_multi_table_selects,
        'isolate_table_selects': isolate_table_selects,
        'optimize_joins': optimize_joins,
        'pushdown_predicates': pushdown_predicates,
        'pushdown_projections': pushdown_projections,
        'qualify_tables': qualify_tables,
        'qualify_columns': qualify_columns,
        'quote_identities': quote_identities,
        'unnest_subqueries': unnest_subqueries,
    }

    @classmethod
    def optimize(cls, optimizer: str, sql: str, schema: Union[dict, None] = None) -> Expression: 
        expr = parse_one(sql, read='spark') 
        optimizer = cls.optimizers[optimizer]
        optimized = optimizer(expr,schema).sql(pretty=True)
        # if optimizer == 'qualify_columns':
        #     optimized = optimizer(expr, schema).sql(pretty=True)
        # else:
        #     optimized = optimizer(expr).sql(pretty=True)
        return optimized

    @classmethod
    def get_optimizers(cls) -> list[str]:
        return cls.optimizers.keys()



from typing import List


class Manifest:
    APP_NAME: str = 'swissql'
    APP_DESCRIPTION: str = 'Spark SQL analytic engine'
    APP_VERSION: str = 'v0.2.2'
    LICENSE: str = 'GNU'
    EMAIL: str = 'andrey24072002@bk.ru'
    REQUIREMENTS : List  = ["lark>=1.1.2", "sqlfluff>=1.2.1", "sqlglot>=4.2.9"]
    AUTHORS_GIT: str = 'husker Nicialy Quakumei'
    GIT_URL: str = 'https://github.com/IPROSpark/SparkSQL-Analyzer'



from typing import Optional
from lark import Lark, Tree
from dataclasses import dataclass

@dataclass
class CustomRule:
    name: str
    grammar: str
    comment: str

    parser: Optional[Lark] = None 

    def __initialise_parser(self) -> None:
        self.parser = Lark(
            self.grammar,
            parser='lalr',
            lexer='basic',
            start='start',
            propagate_positions=True,
        )

    def parse_sql_tree(self, sql: str) -> Tree:
        if not self.parser:
            self.__initialise_parser()
        return self.parser.parse(sql)

    def parse_sql_str(self, sql: str) -> str:
        if not self.parser:
            self.__initialise_parser()
        return repr(self.parser.parse(sql))





import os
import pathlib
from typing import Any, Union
from sqlfluff.cli.commands import cli

class SqlfluffCheck:
    def __init__(self, mode: Union[str,None], file: Union[str,None ] = None, q: Union[str,None] =None, 
                 rules: Union[str,None] =None, dialect: str = None, help: Union[str,None] =None) -> None:
        self.file = file
        self.querry = q
        self.rules = rules
        self.dialect = dialect
        self.mode = mode
        self.help = help
        # TODO пофиксить создание файла
        self.filename = pathlib.Path(__file__).parent.resolve().joinpath(pathlib.Path('intermediate.sql'))
    
    def create_str(self) -> str:
        string = "sqlfluff"
        if self.help:
            return string + ' -' + self.help
        if self.mode:
            string =  string + ' ' + self.mode
        if self.file:
            string =  string + ' ' + self.file
        elif self.querry:
            print(self.filename)
            with open(self.filename, "w") as f:
                f.write(self.querry)
            string = string + ' ' + str(self.filename)
        string = string + ' --dialect ' + self.dialect
        if self.rules:
            string = string + ' --rules' + self.rules
        output = os.popen(string).read()
        if self.querry:
            os.remove(self.filename)
        return output
            
        
